/*
 * Copyright (c) 1993, by Cray Research Superservers, Inc.
 */

#ifndef	_SYS_SCSI_TARGETS_SGADEF_H
#define	_SYS_SCSI_TARGETS_SGADEF_H

#pragma ident	"@(#)sgadef.h 1.5	96/07/10 CRS"

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * Defines for SCSI AutoLoaders
 */

/*
 * Supported AutoLoader types plus default for opening
 */
#define	SA_TYPE_INVALID		0x00	/* invalid type */
#define	SA_TYPE_OTHER		0x01	/* unknown type */
#define	SA_TYPE_EXB		0x02	/* EXB 8mm 210/440/480 */
#define	SA_TYPE_QUANTUM		0x03	/* Quantum DLT 5/7 */
#define	SA_TYPE_FUJI		0x04	/* Fujitsu M2488 */

/*
 * Defines for supported AutoLoader options
 */
#define	SA_PTE			0x0001	/* 0x2b PositionToElt */
#define	SA_PAMR			0x0002	/* 0x1e Prevent/AllowMediumRemoval */
#define	SA_EXCH			0x0004	/* 0xa6 ExchangeMedium */
#define	SA_RVEA			0x0008	/* 0xb5 RequestVolumeElementAddress */
#define	SA_RSVREL		0x0010	/* 0x16/0x17 Reserve/Release */
#define	SA_IES			0x0020	/* 0x07 InitializeElementStatus */

#define	SA_BARCODEREADER	0x0100	/* may have optical bar code reader */
#define	SA_NEEDS_EJECT		0x0200	/* need to eject tape before SGA_MM */
#define	SA_CAN_ROTATE		0x0400	/* can rotate (flip) media */

#define	SA_DYNAMIC		0x1000	/* name is dynamically alloc'ed */

/*
 * Structure for known AutoLoader type table
 */
typedef struct sa_altype {
	char	*sa_name;		/* Name, for debug */
	char	sa_length;		/* Length of vendor id */
	char	sa_vid[24];		/* Vendor id and model (product) id */
	char	sa_type;		/* Drive type for driver */
	int	sa_options;		/* Drive options */
} sa_altype_t;

#if	defined(_KERNEL) || defined(_KMEMUSER)

/*
 * Local definitions, for code clarity
 */
#define	SGA_SCSI_DEVP	(un->un_sd)
#define	SGA_DEVINFO	(SGA_SCSI_DEVP->sd_dev)
#define	SGA_INQUIRY	(SGA_SCSI_DEVP->sd_inq)
#define	SGA_RQSENSE	(SGA_SCSI_DEVP->sd_sense)
#define	SGA_MUTEX	(&SGA_SCSI_DEVP->sd_mutex)
#define	ROUTE		(&SGA_SCSI_DEVP->sd_address)

#define	SCBP(pkt)	((struct scsi_status *)(pkt)->pkt_scbp)
#define	SCBP_C(pkt)	((*(pkt)->pkt_scbp) & STATUS_MASK)
#define	CDBP(pkt)	((union scsi_cdb *)(pkt)->pkt_cdbp)


/*
 * Private info for scsi auto loaders.
 *
 * Pointed to by the un_private pointer of one of the SCSI_DEVICE structures.
 */
typedef struct scsi_al {
	struct scsi_device *un_sd;	/* 0 back pointer to SCSI_DEVICE */
	struct scsi_pkt	*un_rqs;	/* 4 ptr to request sense command pkt */
	struct scsi_pkt	*un_pkt;	/* 8 ptr to current command pkt */
	kcondvar_t	un_pkt_cv;	/* c conditional variable for pkt */
	int		un_pkt_busy;	/* Wait Variable */
	struct	buf	*un_sbufp;	/* for use in special io */
	kcondvar_t	un_sbuf_cv;	/* conditional variable for sbufp */
	int		un_sbuf_busy;	/* Wait Variable */
	int		un_retry_ct;	/* retry count */
	u_int		un_state;	/* current state */
	u_int		un_arq;		/* ARQ mode on this tgt */
	struct buf	*un_rqbp;	/* buf for Request Sense packet */
	struct mode_sense *un_mspl;	/* mode sense info */
	u_char		un_suspended;	/* True, if suspended */
	u_char		un_bcs;		/* 35 True, if bar code scanner avail */
	kcondvar_t	un_cmd_cv;	/* 38 conditional variable for cmd */
	int		un_cmd_busy;	/* Wait Variable */
	struct uscsi_cmd *un_cmd;
	u_char		un_cdb[CDB_GROUP5]; /* max size */

	sa_altype_t	*un_dp;		/* ptr to al table entry */
	u_int		un_dp_size;	/* size of un_dp alloc'ed */
} scsi_al_t;


/*
 * Driver states
 */
#define	SGA_STATE_NIL		0
#define	SGA_STATE_CLOSED	1
#define	SGA_STATE_OPEN		2

/*
 * Parameters
 */
#define	SGA_NAMESIZE	44	/* size of pretty string for vid/pid */
#define	VIDLEN		8	/* size of vendor identifier length */
#define	PIDLEN		16	/* size of product identifier length */


#define	SGA_IO_TIME	60	/* default command timeout, 60 sec */
#define	SGA_UTIMEOUT	20	/* default timeout, 20 sec */

/*
 * 5 seconds is what we'll wait if we get a Busy Status back
 */
#define	SGA_BSY_TIMEOUT		(drv_usectohz(5 * 1000000))

/*
 * Number of times we'll retry a normal operation.
 *
 * This includes retries due to transport failure
 * (need to distinguish between Target and Transport failure)
 */
#define	SGA_RETRY_COUNT		30

/*
 * sga_callback action codes
 */
#define	COMMAND_DONE		0
#define	COMMAND_DONE_ERROR	1
#define	QUE_COMMAND		2
#define	QUE_SENSE		3
#define	JUST_RETURN		4

#endif	/* defined(_KERNEL) || defined(_KMEMUSER) */

struct mode_param_list_head {
	u_char		mode_data_len;	/* mode data length */
	u_char		:8;		/* reserved */
	u_char		:8;		/* reserved */
	u_char		:8;		/* reserved */
};

typedef struct mode_sense {
	struct mode_param_list_head	m_header;
	struct mode_elt_addr_asgn	m_elt_addr_asgn;
} mode_sense_t;


/*
 * These should be added to sun/sys/scsi/generic/commands.h
 */
/*
 * Group 0 Commands, media changer devices
 */
#define	SCMD_INITIALIZE_ELEMENT_STATUS	0x07

/*
 * Group 1 Commands, media changer devices
 */
#define	SCMD_POSITION_TO_ELT		0x2b

/*
 * Group 5 Commands, media changer devices
 */
#define	SCMD_MOVE_MEDIUM		0xa5
#define	SCMD_EXCHANGE_MEDIUM		0xa6
#define	SCMD_READ_ELEMENT_STATUS	0xb8
#define	SCMD_REQUEST_VOLUME_ELT_ADDR	0xb5
#define	SCMD_SEND_VOLUME_TAG		0xb6


typedef struct scsi_pos_to_elt {		/* 2bh position to elt */
	u_char	pte_opcode;
	u_char	pte_lun		:3,		/* logical unit */
				:5;		/* reserved */
	u_char	pte_tea_msb;			/* transport elt addr msb */
	u_char	pte_tea_lsb;			/* transport elt addr lsb */
	u_char	pte_dea_msb;			/* destination elt addr msb */
	u_char	pte_dea_lsb;			/* destination elt addr lsb */
	u_char			:8;		/* reserved */
	u_char			:8;		/* reserved */
	u_char			:7,		/* reserved */
		pte_invert	:1;		/* invert */
} scsi_pos_to_elt_t;

typedef struct scsi_mov_med {			/* a5h move media */
	u_char	mm_opcode;
	u_char	mm_lun		:3,		/* logical unit */
				:5;		/* reserved */
	u_char	mm_tea_msb;			/* transport elt addr msb */
	u_char	mm_tea_lsb;			/* transport elt addr lsb */
	u_char	mm_sea_msb;			/* source elt addr msb */
	u_char	mm_sea_lsb;			/* source elt addr lsb */
	u_char	mm_dea_msb;			/* destination elt addr msb */
	u_char	mm_dea_lsb;			/* destination elt addr lsb */
	u_char			:8;		/* reserved */
	u_char			:8;		/* reserved */
	u_char			:7,		/* reserved */
		mm_invert	:1;		/* invert */
	u_char	mm_control;			/* control */
} scsi_mov_med_t;

typedef struct scsi_exch_med {			/* a6h exchange medium */
	u_char	exm_opcode;
	u_char	exm_lun		:3,		/* logical unit */
				:5;		/* reserved */
	u_char	exm_tea_msb;			/* transport elt addr msb */
	u_char	exm_tea_lsb;			/* transport elt addr lsb */
	u_char	exm_sea_msb;			/* source elt addr msb */
	u_char	exm_sea_lsb;			/* source elt addr lsb */
	u_char	exm_dea_msb;			/* first dest elt addr msb */
	u_char	exm_dea_lsb;			/* first dest elt addr lsb */
	u_char	exm_d2ea_msb;			/* second dest elt addr msb */
	u_char	exm_d2ea_lsb;			/* second dest elt addr lsb */
	u_char			:6,		/* reserved */
		exm_invert2	:1,		/* invert */
		exm_invert1	:1;		/* invert */
	u_char	exm_control;			/* control */
} scsi_exch_med_t;

typedef struct scsi_res {			/* b8h read element status */
	u_char	res_opcode;
	u_char	res_lun		:3,		/* logical unit */
				:1,		/* vol tag */
		res_elt_type	:4;		/* element type code */
	u_char	res_sea_msb;			/* starting elt addr msb */
	u_char	res_sea_lsb;			/* starting elt addr lsb */
	u_char	res_noe_msb;			/* number of elements msb */
	u_char	res_noe_lsb;			/* number of elements lsb */
	u_char			:8;		/* reserved */
	u_char	res_len_msb;			/* allocation length msb */
	u_char	res_len;			/* allocation length */
	u_char	res_len_lsb;			/* allocation length lsb */
	u_char			:8;		/* reserved */
	u_char	res_control;			/* control */
} scsi_res_t;

typedef struct scsi_res_esh {			/* element status data */
	u_char	esd_sea_msb;			/* starting elt addr msb */
	u_char	esd_sea_lsb;			/* starting elt addr lsb */
	u_char	esd_noe_msb;			/* number of elements msb */
	u_char	esd_noe_lsb;			/* number of elements lsb */
	u_char			:3,		/* reserved */
		esd_sac		:5;		/* send action code */
	u_char	esd_bcr_hsb;			/* byte count of report hsb */
	u_char	esd_bcr_msb;			/* byte count of report msb */
	u_char	esd_bcr_lsb;			/* byte count of report lsb */
} scsi_res_esh_t;

typedef struct scsi_res_esp {			/* element status page */
	u_char	esp_elt_type_code;		/* elt type code */
	u_char	esp_pvoltag	:1,
		esp_avoltag	:1,
				:6;		/* reserved */
	u_char	esp_edl_msb;			/* element descrptor length */
	u_char	esp_edl_lsb;			/* element descrptor length */
	u_char			:8;		/* reserved */
	u_char	esp_dda_hsb;			/* byte count desc data msb */
	u_char	esp_dda_msb;			/* byte count desc data sb */
	u_char	esp_dda_lsb;			/* byte count desc data lsb */
} scsi_res_esp_t;

#define	VOL_TAG_LEN	36
typedef struct scsi_res_ed {			/* element status page */
	u_char	ed_ea_msb;			/* elt addr msb */
	u_char	ed_ea_lsb;			/* elt addr lsb */
	u_char			:4,		/* reserved */
		ed_status_access:1,		/* 1=element is accessible */
		ed_status_except:1,		/* 1=cart in abnormal state */
				:1,		/* reserved */
		ed_status_full	:1;		/* 1=elt contains a cartridge */
	u_char			:8;		/* reserved */
	u_char	ed_asc;
	u_char	ed_ascq;
	u_char	ed_notbus	:1,		/* */
				:1,		/* reserved */
		ed_idvalid	:1,		/* 1=valid scsi id */
		ed_luvalied	:1,		/* 1=valid scsi lun */
				:1,		/* reserved */
		ed_scsi_lun	:3;		/* scsi logical unit number */
	u_char	ed_scsi_tgt_id;			/* scsi target id */
	u_char			:8;		/* reserved */
	u_char	ed_svalid	:1,		/* 1=ed_ssea_xsb is valid */
		ed_invert	:1,		/* 1=media inverted */
				:6;		/* reserved */
	u_char	ed_ssea_msb;			/* source elt addr msb */
	u_char	ed_ssea_lsb;			/* source elt addr lsb */
	u_char	ed_pvol_tag[VOL_TAG_LEN];	/* primary volume tag info */
	u_char	reserved[4];			/* reserved */
} scsi_res_ed_t;

typedef struct scsi_snd_vol_tag {		/* b6h send volume tag */
	u_char	svt_opcode;
	u_char	svt_lun		:3,		/* logical unit */
	u_char			:1,		/* reserved */
		svt_elt_type	:4;		/* element type code */
	u_char	svt_ea_msb;			/* transport elt addr msb */
	u_char	svt_ea_lsb;			/* transport elt addr lsb */
	u_char			:8;		/* reserved */
	u_char			:3,		/* vol tag */
		svt_sa_code	:5;		/* Send action code */
	u_char			:8;		/* reserved */
	u_char			:8;		/* reserved */
	u_char	svt_pll_msb;			/* param list length msb */
	u_char	svt_pll_lsb;			/* param list length lsb */
	u_char			:8;		/* reserved */
	u_char	svt_control;			/* control */
} scsi_snd_vol_tag_t;

#ifdef	__cplusplus
}
#endif


#endif	/* _SYS_SCSI_TARGETS_SGADEF_H */
